from golgi.config import configurable
from series.logging import Logging

from amino import IO, Map

from series.api_client import ClientBase, ApiClientMeta, command, ApiClient


@configurable(library_client=['rest_api_port', 'rest_api_url'])
class LibClient(ClientBase, metaclass=ApiClientMeta):

    @property
    def client(self):
        return ApiClient(self._rest_api_url, self._rest_api_port)

    @command('series season episode', 'Create an episode with the supplied metadata')
    def create_episode(self, cmd):
        series, season, episode = cmd.args
        data = dict(episode=episode)
        path = 'series/{}/seasons/{}/episodes'.format(series, season)
        return IO.delay(self.client.post, path, body=data)

    @command('series season episode subfps', 'Set the episode\'s subfps')
    def subfps(self, cmd):
        series, season, episode, subfps = cmd.args
        data = dict(subfps=subfps)
        path = 'series/{}/seasons/{}/episodes/{}'.format(series, season, episode)
        return IO.delay(self.client.put, path, body=data)

    def set_episode_new(self, cmd, new):
        series = cmd.args.head.get_or_fail('no show specified')
        data = cmd.args.lift(1).map(lambda a: Map(season=a)).get_or_strict(Map())
        data1 = cmd.args.lift(2).map(lambda a: data ** Map(episode=a)).get_or_strict(data)
        endpoint = 'new' if new else 'seen'
        path = f'series/{series}/mark_{endpoint}'
        return IO.delay(self.client.put, path, body=data1)

    @command('show season episode[-episode]', 'mark one episode or a range of episodes as seen')
    def mark_episode_seen(self, cmd):
        return self.set_episode_new(cmd, False)

    @command('show season episode[-episode]', 'mark one episode or a range of episodes as new')
    def mark_episode_new(self, cmd):
        return self.set_episode_new(cmd, True)

__all__ = ('LibClient',)



import functools

from toolz import merge

from sqlpharmacy.core import Database

from sqlalchemy import Column, Integer, String, Boolean


def valid(epi: 'Season') -> bool:
    return epi.series is not None


@functools.total_ordering
@Database.many_to_one('Series', backref_name='seasons')
class Season(object, metaclass=Database.DefaultMeta):
    number = Column(Integer, nullable=False)
    subfps = Column(String)
    metadata_fetched = Column(Boolean, default=False)
    metadata_failures = Column(Integer, default=0)

    def __str__(self):
        return '{} season {}'.format(self.series.name, self.number)

    def __repr__(self):
        return '<Season {} ({})>'.format(self.number, self.series.name)

    def __lt__(self, other):
        if isinstance(other, Season):
            return valid(self) and valid(other) and (self.series < other.series or
                                                     (self.series == other.series and
                                                      self.number < other.number))

    def __eq__(self, other):
        if isinstance(other, Season):
            return valid(self) and valid(other) and self.series == other.series and self.number == other.number

    def __hash__(self):
        return hash((self.series, self.number))

    @property
    def info(self):
        return dict(
            id=self.id,
            series=self.series.formatted_name,
            number=self.number,
            new_count=self.new_count,
            subfps=self.subfps,
        )

    @property
    def ext_info(self):
        ext = dict(show=self.series.info)
        return merge(self.info, ext)

    @property
    def new_count(self):
        return len([e for e in self.episodes if e.new])

    @property
    def empty(self):
        return (all(e.removed for e in self.episodes) or
                self.episodes.count == 0)

    @property
    def subfps_fallback(self):
        return self.subfps or self.series.subfps

__all__ = ['Season']



from datetime import datetime

from series.get.handler import ShowHandler, S
from series.condition import LambdaCondition
from series.tvdb.data import TvdbShow, TvdbEpisode
from series.get import Show
from series.tvdb.api import show_by_id, search_show, current_episodes

from golgi.config import configurable

from amino import __, IO, do, Do, Maybe


@configurable(series=['monitor'], show_planner=['check_interval'])
class ShowPlanner(ShowHandler):
    ''' Queries a tv db for episode schedules.
    If a show currently has no next episode, e.g. because it has just
    aired, the tv db service is queried once a day until there is an
    airdate.
    '''

    def __init__(self, releases, shows, *a, **kw):
        super().__init__(shows, 1, 'show planner', *a, **kw)
        self._releases = releases
        self._shows_initialized = False

    def _sanity_check(self):
        if not self._shows_initialized:
            self._init()

    def _init(self):
        for name in self._monitor:
            self._init_show(name)
        self._shows_initialized = True

    def _init_show(self, name):
        if not self._shows.name_exists(name):
            self._add_show(name)

    def _add_show(self, name: str) -> None:
        self.log.debug('Adding show for "{}"'.format(name))
        tvdb_show = search_show(name).attempt.join.to_maybe.flat_map(lambda a: a.head)
        def add(s: TvdbShow) -> None:
            show = self._shows.add(name, s)
            self._update_show(show, s)
        tvdb_show.cata(add, lambda: self.log.debug('Adding show failed.'))

    def _handle(self, show):
        self._handle_io(show).fatal

    @do(IO[None])
    def _handle_io(self, show: Show) -> Do:
        show.last_check = datetime.now()
        tvdb_show = yield show_by_id(show.etvdb_id)
        yield tvdb_show.cata(
            IO.delay(lambda e: self.log.error(f'Show `{show.name}` couldn\'t be found anymore: {e}')),
            lambda t: self._update_show(show, t),
        )

    @do(IO[None])
    def _update_show(self, show: Show, tvdb_show: TvdbShow) -> Do:
        self.log.debug('Updating show {}'.format(tvdb_show.name))
        data = {}
        if tvdb_show.ended:
            data['ended'] = True
            self.log.warn('Show has ended: {}'.format(show.name))
        else:
            data = yield self._fetch_next_episode(show, tvdb_show)
        self._shows.update(show.id, data)

    @do(IO[dict])
    def _fetch_next_episode(self, show: Show, tvdb_show: TvdbShow) -> Do:
        data = {}
        self.log.debug('Fetching next episode for "{}"'.format(show.name))
        current = yield current_episodes(tvdb_show.id)
        def handle_current(latest: Maybe[TvdbEpisode], next: Maybe[TvdbEpisode]) -> None:
            latest.foreach(lambda a: data.update(latest_episode=a.number, latest_season=a.season))
            def set_date(epi: TvdbEpisode, date: int) -> None:
                data['next_episode_stamp'] = date
                data['season'], data['next_episode'] = epi.season, epi.number
                self.log.debug(f'Found season {epi.season} episode {epi.number}')
            (next & next.flat_map(lambda a: a.airdate)).map2(set_date).get_or(lambda: self.log.debug('no next episode'))
        current.map2(handle_current).get_or(lambda e: self.log.debug(f'Fetching next episode failed: {e}.'))
        return data

    @property
    def _conditions(self):
        return (
            ~S('has_next_episode') & ~S('ended') &
            LambdaCondition('recheck interval',
                            __.can_recheck(self._check_interval))
        )

    def activate_id(self, id):
        super().activate_id(id)
        self._shows.update_by_id(id, last_check_stamp=0)

__all__ = ['ShowPlanner']



from golgi.config import configurable

from tek_utils.sharehoster.torrent import torrent_cacher_client
from tek_utils.sharehoster.errors import ShareHosterError

from amino import List

from series.get.handler import ReleaseHandler, R, Li
from series.get.errors import SeriesException
from series.condition import HandlerCondition, DynOr
from series.get.model.release import ReleaseMonitor
from series.get.model.link import Torrent


class CachableTorrent(HandlerCondition):

    def __init__(self, torrent: Torrent) -> None:
        self.torrent = torrent

    @property
    def cond(self):
        return ~(Li('caching') | Li('valid') | Li('dead'))

    def ev(self, release: ReleaseMonitor):
        return self.cond.ev(self.torrent)

    def describe(self, item, target):
        d = self.cond.describe(self.torrent, target)
        m = self.torrent.name
        return '{} -> {}'.format(m, d)


class CachableTorrents(DynOr):

    @property
    def _sub_type(self):
        return CachableTorrent

    def _dyn_subs(self, item: ReleaseMonitor):
        return List.wrap(item.torrent_links)

    def _multiline(self, sub) -> bool:
        return True

    def describe(self, item, target):
        d = super().describe(item, target)
        return 'cachable torrents => {}'.format(d)


@configurable(torrent=['cacher'])
class TorrentHandler(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'torrent handler', **kw)
        self._client = torrent_cacher_client()

    def _handle(self, monitor):
        for link in monitor.cachable_torrents:
            msg = 'Requesting torrent download for {}…'.format(monitor.release)
            self.log.info(msg)
            try:
                link.torrent.request()
                self._releases.update_link(link.id, caching=True)
            except ShareHosterError as e:
                msg = 'Error requesting torrent for {}: {}'
                self.log.error(msg.format(monitor.release, e))

    def _cleanup(self):
        self._check_error()

    def _check_error(self):
        try:
            errors = [t.get('id', 0) for t in self._client.transfers if t.get('status') == 'ERROR']
            if errors:
                self.log.info('Canceling erroneous torrents.')
                self._client.cancel_transfers(errors)
                self._client.clean_transfers()
        except ShareHosterError as e:
            self.log.error(f'connection error while checking torrents: {e}')

    @property
    def _conditions(self):
        return ~R('downloaded') & CachableTorrents()

    def _qualify(self, monitor):
        try:
            return super()._qualify(monitor)
        except ShareHosterError as e:
            msg = 'Error checking torrent status for {}: {}'
            self.log.error(msg.format(monitor.release, e))

    @property
    def _candidates(self):
        return self._no_cached_torrents

    def _sanity_check(self):
        self._check_cacher_config()
        self._check_service_accessible()

    def _check_cacher_config(self):
        if not self._cacher:
            raise SeriesException('No torrent cacher configured!')

    def _check_service_accessible(self):
        cacher = torrent_cacher_client()
        if not cacher.account_info:
            raise SeriesException(
                'Couldn\'t access torrent cacher \'{}\'!'.format(self._cacher))

__all__ = ['TorrentHandler']



import operator
from datetime import datetime, timedelta

from amino import IO, Maybe, Lists, do, Do, Either
from amino.json.data import JsonError

from series.get.handler import ShowHandler

from series.condition import HandlerCondition, SimpleCondition, LambdaCondition
from series.util import unix_to_datetime
from series.get import Show
from series.tvdb.api import airdate


class AirsToday(HandlerCondition):

    @property
    def _today_thresh(self):
        return datetime.now() - timedelta(hours=16)

    def _check(self, show):
        return show.next_episode_date < self._today_thresh

    def ev(self, show):
        return show.has_next_episode and self._check(show)

    def describe(self, show, target):
        match = self.ev(show)
        good = match == target
        today = (show.next_episode_date > self._today_thresh) and match
        desc = ('today' if today else
                show.next_episode_date.strftime('%F') if
                show.next_episode_stamp > 0 else
                'no date')
        return 'airs today[{}]'.format(self._paint(desc, good))


class CanCatchUp(SimpleCondition):
    ''' *current* is the last aired episode, which includes those aired
    within 24 hours in the future.
    *latest* is the newest release in the db.
    '''

    def __init__(self, latest) -> None:
        self._latest = latest

    def latest_episode(self, show):
        return show.latest_episode_m

    def latest_release(self, show):
        return self._latest(show)

    def ev(self, show):
        return (self.latest_episode(show) & self.latest_release(show)
                ).map2(operator.gt).true

    @property
    def _desc(self):
        return 'latest > downloaded'

    def _repr(self, show, match):
        op = '>' if match else '<'
        def has(release):
            return '{} {} {}'.format(self.latest_episode(show) | -1, op,
                                     release)
        return self.latest_release(show) / has | 'no latest episode'


class ShowScheduler(ShowHandler):

    def __init__(self, releases, shows, **kw):
        super().__init__(shows, 60, 'show scheduler', cooldown=3600, **kw)
        self._releases = releases

    def _handle(self, show: Show) -> None:
        self._handle_io(show).fatal

    @do(IO[None])
    def _handle_io(self, show: Show) -> Do:
        self.log.debug(f'attempting to schedule {show}')
        response = yield airdate(show.etvdb_id, show.season, show.next_episode)
        def error(e: JsonError) -> IO[None]:
            IO.delay(self.log.error, f'failed to parse repsonse for airdate of next episode of `{show}`: {e}')
        def present(ad: int) -> IO[None]:
            return (
                self._handle_invalid_show(show, ad)
                if show.has_next_episode and ad != show.next_episode_stamp else
                self._handle_valid_show(show)
            )
        def absent() -> IO[None]:
            return self._handle_valid_show(show)
        def success(ad: Maybe[int]) -> IO[None]:
            return ad.cata(present, absent)
        yield response.cata(error, success)

    def _handle_valid_show(self, show: Show) -> IO[None]:
        def catch_up(latest_release, latest_episode):
            return (
                Lists.range(latest_release + 1, latest_episode + 1)
                .traverse(lambda e: self._schedule(show, show.latest_season, e), IO)
            )
        def schedule_next():
            return self._schedule(show, show.season, show.next_episode)
        return (
            schedule_next()
            if AirsToday().ev(show) else
            (self._latest(show) & show.latest_episode_m).map2(catch_up).get_or_strict(IO.pure(None))
        )

    @do(IO[None])
    def _handle_invalid_show(self, show: Show, airdate: int) -> Do:
        new_date = unix_to_datetime(airdate).date()
        msg = f'{show.name} {show.season}x{show.next_episode} was rescheduled: {show.next_episode_day} => {new_date}'
        yield IO.delay(self.log.error, msg)
        yield IO.delay(self._update, show, next_episode_stamp=airdate)

    @property
    def _conditions(self):
        return LambdaCondition('next has no release', self._no_next_release) & (AirsToday() | CanCatchUp(self._latest))

    def _latest(self, show):
        return self._releases.latest_for_season(show.canonical_name, show.latest_season)

    def _no_next_release(self, show):
        return not self._latest(show).contains(show.next_episode)

    @do(IO[None])
    def _schedule(self, show: Show, season: int, episode: int) -> Do:
        response = yield airdate(show.etvdb_id, season, episode)
        @do(Either[str, None])
        def schedule() -> Do:
            data = yield response
            airdate = yield data.to_either(lambda: f'no airdate')
            msg = f'Scheduling release "{show.name} {season}x{episode}" on {unix_to_datetime(airdate)}'
            self.log.info(msg)
            self._releases.create(show.canonical_name, season, episode, airdate,
                                  downgrade_after=show.downgrade_after,
                                  search_name=show.search_name)
            if show.season > show.latest_season:
                self._shows.update_by_id(show.id, latest_season=show.season)
            self._commit()
        result = yield IO.delay(schedule)
        yield result.cata(
            lambda e: IO.delay(lambda: self.log.error(f'couldn\'t schedule {show} {season}x{episode}: {e}')),
            IO.pure
        )


__all__ = ['ShowScheduler']



# from pathlib import Path

from setuptools import setup, find_packages

version_parts = (2, 35, 26)
version = '.'.join(map(str, version_parts))

# try:
#     from golgi.config.write import write_pkg_config
# except ImportError:
#     conf = []  # type: ignore
# else:
#     write_pkg_config('.', Path('series.conf'), 'series')
#     conf = ['series.conf']

setup(
    name='series',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='GPLv3',
    packages=find_packages(exclude=['integration', 'integration.*', 'unit',
                                    'unit.*']),
    # data_files=[('share/series/config', conf)],
    package_data={
        '': ['alembic/env.py', 'alembic.ini', 'alembic/versions/*.py'],
    },
    install_requires=[
        'amino~=13.0.1',
        'golgi~=2.0.1',
        'tek-utils~=4.0.1',
        'tek~=4.1.1',
        'requests==2.19.1',
        'sqlpharmacy==3.0.0',
        'lxml==4.2.3',
        'flask==0.12.2',
        'alembic==0.9.10',
        'guessit==3.1.0',
    ],
    entry_points={
        'console_scripts': [
            'serieslibd = series.library.cli:libd',
            'serieslibc = series.library.cli:libc',
            'seriesgetd = series.get.cli:getd',
            'seriesgetc = series.get.cli:getc',
            'subsync_dir = series.subsync:subsync_dir_cli',
            'subsync_cwd = series.subsync:subsync_cwd',
            'subsync_auto = series.subsync:subsync_auto',
            'handle_episode = series.handle_episode:handle_episode_cli',
            'store_episode = series.store_episode:store_episode_cli',
        ]
    }
)



import abc
import time
from threading import Thread
from datetime import datetime, timedelta

import sqlalchemy

from series.logging import Logging

import amino
from amino import List, Map, __

from series.errors import SeriesDException
from series.condition import HandlerCondition, StrictCondition


class Job:

    def __init__(self, item) -> None:
        self.item = item


class Handler(Thread, Logging, metaclass=abc.ABCMeta):

    def __init__(self, interval, description, cooldown=0, inactive_interval=60):
        self._interval = 1 if amino.development else interval
        self._description = description
        self._inactive_interval = 1 if amino.development else inactive_interval
        super().__init__(name=self._description)
        self._running = True
        self._initial_wait = 5
        self._failures = 0
        self._cooldown = 0 if amino.development else cooldown
        self._last_check = {}
        self._failure_limit = 20

    def run(self):
        self.log.info('Starting {}.'.format(self._description))
        try:
            self._sanity_check()
        except SeriesDException as e:
            self.log.error('Shutting down {} with reason: {}'.format(
                self._description, e))
        else:
            self._main_loop()

    def _main_loop(self):
        next_check = datetime.now() + timedelta(seconds=self._initial_wait)
        while self._running:
            if datetime.now() > next_check:
                next_check = (self._next_check if self._check_error_wrapper()
                              else self._next_check_inactive)
            time.sleep(1)

    def stop(self):
        self._running = False

    def _check_error_wrapper(self):
        ''' *handled* is indicative of whether a candidate was found and
        processed.
        Returns *True* if there are remaining jobs in the database.
        '''
        handled = False
        try:
            handled = self._check()
        except sqlalchemy.exc.InvalidRequestError as e:
            self.log.debug(e)
        except Exception as e:
            self._failure(e)
        else:
            if handled:
                self._success()
            return handled
        return True

    def _check(self):
        with self._lock:
            ret = self._current.effect(self._process).is_just
            self._cleanup()
            return ret

    def _process(self, job):
        item = job.item
        if hasattr(item, 'id'):
            self._last_check[str(item.id)] = time.time()
        self._handle_job(job)

    def _cleanup(self):
        pass

    @property
    def _next_check(self):
        return datetime.now() + timedelta(seconds=self._interval)

    @property
    def _next_check_inactive(self):
        return datetime.now() + timedelta(seconds=self._inactive_interval)

    def _handle_job(self, job):
        return self._handle(job.item)

    def _handle(self, task):
        pass

    @property
    def _current(self):
        return self._choose(self._qualified)

    def _choose(self, qualified):
        return qualified.head

    def _cool(self, job):
        item = job.item
        return (not hasattr(item, 'id') or
                self._time_since_last(item) > self._cooldown)

    def _time_since_last(self, item):
        return time.time() - self._last_check.get(str(item.id), 0)

    def activate_id(self, id):
        if str(id) in self._last_check:
            del self._last_check[str(id)]

    @abc.abstractproperty
    def _candidates(self):
        ...

    def _qualify(self, candidate):
        return self._conditions.ev(candidate)

    def _qualify_job(self, job):
        return (self._job_conditions.lift(type(job)) / __.ev(job.item) |
                (lambda: self._qualify(job.item)))

    @property
    def _qualified(self):
        def job(item):
            return item if isinstance(item, Job) else Job(item)
        return (
            List.wrap(self._candidates)
            .map(job)
            .filter(self._cool)
            .filter(self._qualify_job)
        )

    def explain(self, item):
        desc = __.describe(item, target=True)
        cond = (
            desc(self._conditions)
            if self._job_conditions.empty else
            (self._job_conditions.v / desc).join_lines
        )
        return dict(name=self._description, cond=cond)

    @property
    def _conditions(self) -> HandlerCondition:
        return StrictCondition(False)

    @property
    def _job_conditions(self):
        return Map()

    def _commit(self):
        pass

    def _sanity_check(self):
        pass

    def _failure(self, exc):
        self._failures += 1
        import traceback
        self.log.error('Error in {}:'.format(self._description))
        self.log.error(exc)
        traceback.print_exc()
        if amino.development or self._failures >= self._failure_limit:
            msg = f'{self._failure_limit} failures in a row in {self._description}, shutting down.'
            self.log.error(msg)
            self.stop()

    def _success(self):
        self._failures = 0

    @property
    def _lock(self):
        raise NotImplementedError('Handler._lock')

__all__ = ['Handler']



from sqlalchemy.exc import OperationalError

from series.db import Database as Base, FileDatabase as FileBase
from series.get.model.release import ReleaseMonitor
from series.get.model.link import Link


class Database(Base):

    def __init__(self, *a, **kw):
        super().__init__('series.get', *a, **kw)


class FileDatabase(FileBase):

    def __init__(self, *a, **kw):
        super().__init__('series.get', *a, **kw)
        self._reset_ephemeral_attributes()
        self._delete_broken_monitors()
        self.commit()

    def _reset_ephemeral_attributes(self):
        try:
            for release in self.query(ReleaseMonitor):
                release.downloading = False
        except OperationalError:
            pass

    def _delete_broken_monitors(self):
        try:
            for monitor in self.query(ReleaseMonitor):
                if monitor.release is None:
                    self.delete(monitor)
        except OperationalError:
            pass

__all__ = ['FileDatabase', 'Database']



import re
import os
from io import StringIO
from urllib.parse import urljoin
from operator import attrgetter

from lxml import etree

import requests

from golgi import Config, cli
from golgi.config import configurable, ConfigError

from tek.user_input import CheckboxList
from tek.tools import parallel_map

from amino import Path, Lists

from series import SubtitleMetadata, latest_season_dir
from series import dir_info, episode_number, get_release

from series.subsync.errors import NoSubsForEpisode
from series.logging import Logging, log


@configurable(subsync=['series_url_map'])
class SubtitleQueryData(SubtitleMetadata):

    def __init__(self, search_name, canonical_name, season, episode,
                 release=None, extension='srt'):
        self.canonical_name = canonical_name
        super().__init__(search_name, season, episode, release=release,
                         extension=extension)

    @property
    def local_path(self):
        path = Path(self.canonical_name) / 's{}'.format(self.season)
        if self.extension == 'srt':
            path = path / 'sub'
        return path / self.filename

    @property
    def file_series_name(self):
        return self.canonical_name

    @property
    def query_data(self):
        return [self._series_name] + self[1:]

    @property
    def _series_name(self):
        return self._series_url_map.get(self[0], self[0])


@configurable(subsync=['base_url', 'cookies'])
class SubtitleHTTPHandler(object):

    def get(self, url):
        if self._base_url is None:
            raise ConfigError('subsync: config option "base_url" must be set!')
        url = urljoin(self._base_url, url)
        return requests.get(url, cookies=self._cookies).text


def sanitize_sub(sub: str) -> str:
    return Lists.lines(sub).drop_while(lambda a: a.startswith('<')).join_lines


@configurable(series=['series_dir'], subsync=['base_url', 'cookies'])
class Subtitle(Logging):

    def __init__(self, info, release, url, hearing_impaired):
        self.metadata = info
        self.release = release
        self._url = url
        self._content = None
        self.hearing_impaired = hearing_impaired
        self._http = SubtitleHTTPHandler()

    @property
    def local_path(self):
        return self.metadata.local_path

    @property
    def content(self):
        if self._content is None:
            try:
                self._content = sanitize_sub(self._http.get(self._url))
            except requests.RequestException as e:
                self.log.error('Subtitle: {}'.format(e))
        return self._content

    def write(self):
        self.write_to(self.local_path)

    def write_to(self, local_path):
        path = self._series_dir / local_path
        base = path.parent
        if not base.is_dir():
            base.mkdir(parents=True, exist_ok=True)
        with path.open('w') as f:
            if self.content:
                f.write(self.content)
            else:
                self.log.error('Could not download the subtitle.')


class HTMLParser(object):
    xpath_deaf = "descendant::img[@title='Hearing Impaired']"
    xpath_download = "descendant::a[@class='buttonDownload']"
    release_re = re.compile(r'Version ([^,]+),')
    parser = etree.HTMLParser()

    def __init__(self, info):
        self._info = info
        self._http = SubtitleHTTPHandler()

    def subtitles(self):
        url = 'serie/{}/{}/{}/1'.format(*self._info.query_data)
        html = str(self._http.get(url))
        return self._parse(html)

    def _parse(self, html):
        tree = etree.parse(StringIO(html), self.parser)
        if tree.getroot() is None:
            raise NoSubsForEpisode(*self._info)
        epi_divs = tree.xpath("//div[@id='container95m']")
        return [_f for _f in map(self.subtitle_from_div, epi_divs) if _f]

    def subtitle_from_div(self, div):

        def release():
            for line in div.itertext():
                match = self.release_re.search(line)
                if match:
                    return match.group(1).lower()

        def subtitle():
            download = div.xpath(self.xpath_download)[-1]
            return download.attrib['href']
        try:
            hearing_impaired = bool(div.xpath(self.xpath_deaf))
            return Subtitle(self._info, release(), subtitle(),
                            hearing_impaired)
        except IndexError:
            pass


def get_episode(search_name, season, number, release=None,
                canonical_name=None):
    from tek.tools import find
    canon = canonical_name or search_name
    info = SubtitleQueryData(search_name, canon, season, number, release)
    parser = HTMLParser(info)
    subs = parser.subtitles()
    if not subs:
        raise NoSubsForEpisode(search_name, season, number)
    same_release = lambda sub: sub.release == release
    if release is not None:
        matching = list(filter(same_release, subs))
        if matching:
            subs = matching
    not_deaf = lambda sub: not sub.hearing_impaired
    return find(not_deaf, subs) or subs[0]


def write_subs(subs):
    input = CheckboxList(list(map(attrgetter('local_path'), subs)),
                         [True] * len(subs),
                         text_pre=['Downloading these subtitles:'],
                         text_post=['Write to series dir?'])
    write = input.read()

    def process(args):
        sub, doit = args
        if doit:
            sub.write()

    return parallel_map(process, zip(subs, write))


def subsync_episode(metadata, write=True):
    try:
        sub = get_episode(*metadata.all)
        if write:
            sub.write()
        return sub
    except NoSubsForEpisode:
        pass


def target_episodes(vid_epis, sub_epis):
    conf = Config['subsync']
    targets = conf.episodes
    if not targets:
        if conf.only_latest:
            vid_max = max(vid_epis) if vid_epis else 0
            sub_max = max(sub_epis) if sub_epis else 0
            targets = range(sub_max + 1, vid_max + 1)
        else:
            targets = set(vid_epis) - set(sub_epis)
    return sorted(targets)


def subsync_dir(_dir, write=True):
    subs = []
    series_name, season = dir_info(_dir)
    if season is None:
        log.error('Not a season dir!')
    else:
        subs_dir = os.path.join(_dir, 'sub')
        if not os.path.isdir(subs_dir):
            os.makedirs(subs_dir)
        present_vids = os.listdir(_dir)
        present_subs = os.listdir(subs_dir)
        vid_epis = [_f for _f in map(episode_number, present_vids) if _f]
        sub_epis = [_f for _f in map(episode_number, present_subs) if _f]
        targets = target_episodes(vid_epis, sub_epis)

        def process(number):
            try:
                release = get_release(series_name, season, number)
                return get_episode(series_name, season, number, release)
            except NoSubsForEpisode:
                log.warning('No subs found for episode {}!'.format(number))

        subs = [sub for sub in parallel_map(process, targets) if sub is not
                None]
        if subs and write:
            write_subs(subs)
    return subs


@cli(positional=(('cli_dir', 1), ('episodes', '*')))
def subsync_dir_cli():
    _dir = Config['subsync'].cli_dir[0]
    return subsync_dir(_dir)


@cli(positional=(('episodes', '*'),))
def subsync_cwd():
    return subsync_dir(os.getcwd())


@cli()
def subsync_auto():
    monitor = Config['series'].monitor

    def sync():
        for _dir in map(latest_season_dir, monitor):
            yield subsync_dir(_dir, write=False)

    subs = sum(sync(), [])
    if subs:
        write_subs(subs)
